MODULE Sortieren EXPORTS Main;

(* Dieses Programm sortiert beliebig viele, eingegebene positive Zahlen
   unter Verwendung eines Binrbaums. 

   Autor    : Moritz Schnizler, RWTH Aachen
   Umgebung : PM 3, Windows NT 4.0
   Erstellt : 04.12.00  Letzte Aenderung: 05.12.00 *)


IMPORT SIO; (* Importiere Operationen fuer die Ein-/Ausgabe *)


(* Notwendige Datenstruktur fuer den Binaerbaum *)

TYPE KnotenRef = REF Knoten;

     Knoten = RECORD
                wert: CARDINAL;
                lKnoten, rKnoten: KnotenRef;
              END;


PROCEDURE Einfuegen(VAR baum: KnotenRef; wert: CARDINAL) =
(* Fuegt einen neuen Knoten fr den gegebenen CARDINAL-Wert in den Baum ein *)
BEGIN
  IF baum = NIL THEN
    (* Ende des Zweigs erreicht, lege neuen Knoten an *)
    baum := NEW(KnotenRef);

    baum^.wert := wert;
    baum^.lKnoten := NIL;
    baum^.rKnoten := NIL;
  ELSE

    IF wert < baum^.wert THEN
      (* Kleineren Wert in linken Zweig einfuegen *)
      Einfuegen(baum^.lKnoten, wert);
    ELSE
      (* Ansonsten in rechten Zweig einfuegen *)
      Einfuegen(baum^.rKnoten, wert);
    END;

  END;
END Einfuegen;


PROCEDURE Ausgeben(baum: KnotenRef) =
(* Gibt die gespeicherten Werte aufsteigend sortiert aus *)
BEGIN
  (* Abbruchkriterium: Knoten leer! *)
  IF baum # NIL THEN
    (* Erst linken Zweig ausgeben ... *)
    Ausgeben(baum^.lKnoten);

    (* ... dann den eigenen Wert ... *) 
    SIO.PutText(" "); SIO.PutInt(baum^.wert);

    (* ... und schliesslich den rechten Zweig *)
    Ausgeben(baum^.rKnoten); 
  END;
END Ausgeben;


VAR wurzel   : KnotenRef := NIL; (* Wurzel des Baums *)
    
    (* Hilfsvariablen fuer Benutzerinteraktion *)
    zahl     : INTEGER;
    restzeile: TEXT;

BEGIN
  SIO.PutLine("Geben Sie die zu sortierenden, positiven Zahlen ein!");
  SIO.PutLine("(Abbruch mit negativer Zahl)");

  REPEAT
    SIO.PutText("Zahl: ");
    zahl := SIO.GetInt();
    restzeile := SIO.GetLine();

    IF zahl >= 0 THEN
      Einfuegen(wurzel, zahl);
    END;
  UNTIL zahl < 0;
 
  Ausgeben(wurzel); 
  SIO.Nl();

END Sortieren.














































