MODULE InstMeldeamt EXPORTS Main;

(* 
   Dieses Programm realisiert eine einfache Verwaltung fuer Einwohnermeldedaten.
   Diese Version wurde instrumentiert, um den Grad der Zweigueberdeckung zu
   ermitteln.

   Autor           : Moritz Schnizler, RWTH Aachen
   Umgebung        : PM 3, Windows 2000
   Erstellt        : 23.11.00
   Letzte Aenderung: 19.12.00
*)

IMPORT SIO;  (* Importiere notwendige Ein-/Ausgabeoperationen *)
IMPORT Text; (* Importiere notwendige Operationen fr Typ TEXT *)

CONST MAXEINWOHNER = 10; (* Maximale Anzahl der zu verwaltenden Einwohner*)

TYPE (* Typ Person umfasst alle Namen einer Person *)  
  
     Person         = RECORD
                        familienname : TEXT;
                        rufname      : TEXT;
                        vorname      : ARRAY [2..5] OF TEXT;
                      END;


     (* Notwendige Unterbereichstypen und Typ fuer Tagesdatum *)

     Kalendertag    = [1..31];
     Kalendermonat  = [1..12];
     Kalenderjahr   = [1800..3000];   (* Zeitraum ausreichend fuer Programmeinsatz *)         

     Tagesdatum     = RECORD
                        tag  : Kalendertag;
                        monat: Kalendermonat;
                        jahr : Kalenderjahr;
                      END;


     (* Typ Geburtsdaten zusammengesetzt aus Geburtsort und -datum *)

     Geburtsdaten   = RECORD
                        ort  : TEXT;
                        datum: Tagesdatum;
                      END;


     (* Typ Geschlecht als Aufzaehlungstyp *)

     Geschlecht     = {maennlich, weiblich};


     (* Notwendige Unterbereichstypen und Typ fuer Adresse *)

     Hausnummer     = [1..1000];
     Postleitzahl   = [0..99999];

     Adresse        = RECORD
                        strasse: TEXT;
                        nummer : Hausnummer;
                        zusatz : TEXT;         (* Appartment, Block, ... *)
                        plz    : Postleitzahl;
                        ort    : TEXT;
                      END;


     (* Typ Familienstand als Aufzaehlungstyp der vier Moeglichkeiten *)

     Familienstand  = {ledig, verheiratet, geschieden, verwitwet};


     (* Typ Einwohner beschreibt einen Gemeldeten komplett *)

     Einwohner      = RECORD
                        person       : Person;
                        geboren      : Geburtsdaten;
                        geschlecht   : Geschlecht;
                        wohnung      : Adresse;
                        familienstand: Familienstand;
                        ehepartner   : Person;
                        anmeldung    : Tagesdatum;
                      END;


     (* Typ Melderegister umfasst MAXEINWOHNER gemeldete Einwohner *)

     Melderegister  = ARRAY [1..MAXEINWOHNER] OF Einwohner;


(* Notwendige Daten und Prozeduren fr Instrumentierung *)

VAR zaehler := ARRAY [1..67] OF CARDINAL {0, ..};


PROCEDURE ZaehleZweig(i: CARDINAL) =
BEGIN
  INC(zaehler[i]);
END ZaehleZweig; 


PROCEDURE ZaehlerAuswerten() =
VAR anzahl: CARDINAL := 0;

BEGIN
  (* Zaehler ausgeben und
     gleichzeitig Anzahl der ausgefuehrten Zweige zaehlen *)
  FOR i := FIRST(zaehler) TO LAST(zaehler) DO
    SIO.PutText("Z");
    IF i < 10 THEN
      SIO.PutChar('0');
    END;
    SIO.PutInt(i); SIO.PutText(": "); SIO.PutInt(zaehler[i]);
    SIO.PutText("  ");
    IF i MOD 10 = 0 THEN
      SIO.Nl();  (* Nach zehn Zaehlern neue Zeile anfangen! *)
    END;

    IF zaehler[i] > 0 THEN
      anzahl := anzahl + 1; (* Zweig wurde mindestens einmal besucht *)
    END;
  END;

  SIO.Nl(); SIO.Nl();
   
  (* Berechne den Grad der Zweigueberdeckung in Prozent *)
  SIO.PutReal(FLOAT(anzahl, REAL) / FLOAT(LAST(zaehler), REAL));
  SIO.PutLine(" % der Programmzweige wurden ueberdeckt.");
END ZaehlerAuswerten;


(* Prozeduren fr die Ein- und Ausgabe der Datenstrukturelemente *)

PROCEDURE PersonEingeben(): Person =
(* Eingeben aller Namen einer Person *)
VAR ergPerson: Person;
    i        : INTEGER;

BEGIN
  ZaehleZweig(1);
  SIO.PutText("Familienname: ");
  ergPerson.familienname := SIO.GetLine();
  SIO.PutText("Rufname: ");
  ergPerson.rufname := SIO.GetLine();
  
  (* Eingabe von bis zu vier weiteren Vornamen.
     Abbruch, falls vier Namen oder nichts eingegeben wurde. *)
  i := FIRST(ergPerson.vorname) - 1; (* Minus 1, da zuerst inkrementiert wird *)
  
  REPEAT
    ZaehleZweig(2);
    INC(i); (* Eingabe des naechsten Vornamens *)

    (* Drucke den Text "i. Vorname: " *)
    SIO.PutText(Text.FromChar( VAL( ORD('0') + i, CHAR) ) & ". Vorname: ");
    ergPerson.vorname[i] :=SIO.GetLine();
  UNTIL (i = LAST(ergPerson.vorname)) OR Text.Equal(ergPerson.vorname[i], ""); 
  ZaehleZweig(3);

  RETURN ergPerson;
END PersonEingeben;


PROCEDURE PersonAusgeben(p: Person) =
(* Ausgeben aller Namen einer Person *)
VAR i: INTEGER;

BEGIN
  ZaehleZweig(4);
  SIO.PutText(p.familienname);
  SIO.PutText(", ");
  SIO.PutText(p.rufname);
  
  (* Weitere Vornamen ausgeben, solange maximale Anzahl
     nicht erreicht und kein leerer Name angetroffen wurde. *)
  i := FIRST(p.vorname);

  WHILE (i <= LAST(p.vorname)) AND NOT Text.Equal(p.vorname[i], "") DO
    ZaehleZweig(5);
    SIO.PutText(" ");
    SIO.PutText(p.vorname[i]);
    INC(i);
  END;
  ZaehleZweig(6);
END PersonAusgeben;


PROCEDURE DatumEingeben(): Tagesdatum =
(* Eingeben eines vollstaendigen Datums *)
VAR ergDatum: Tagesdatum;

BEGIN
  ZaehleZweig(7);
  SIO.PutText("Tag: ");
  ergDatum.tag   := UBEingeben(FIRST(Kalendertag), LAST(Kalendertag));
  SIO.PutText("Monat: ");
  ergDatum.monat := UBEingeben(FIRST(Kalendermonat), LAST(Kalendermonat));
  SIO.PutText("Jahr: ");
  ergDatum.jahr  := UBEingeben(FIRST(Kalenderjahr), LAST(Kalenderjahr));

  RETURN ergDatum;
END DatumEingeben;


PROCEDURE DatumAusgeben(dat: Tagesdatum) =
(* Ausgeben eines vollstaendigen Datums *)

BEGIN
  ZaehleZweig(8);
  SIO.PutInt(dat.tag); SIO.PutText(".");
  SIO.PutInt(dat.monat); SIO.PutText(".");
  SIO.PutInt(dat.jahr);
END DatumAusgeben;


PROCEDURE GeburtsdatenEingeben(): Geburtsdaten =
(* Eingeben der kompletten Geburtsdaten *)
VAR ergGeburtsdaten: Geburtsdaten;

BEGIN
  ZaehleZweig(9);
  SIO.PutText("Geburtsort: ");
  ergGeburtsdaten.ort   := SIO.GetLine();
  SIO.PutLine("Geburtsdatum: ");
  ergGeburtsdaten.datum := DatumEingeben();

  RETURN ergGeburtsdaten;
END GeburtsdatenEingeben;


PROCEDURE GeburtsdatenAusgeben(gd: Geburtsdaten) =
(* Ausgeben der kompletten Geburtsdaten *)

BEGIN
  ZaehleZweig(10);
  SIO.PutText("Geboren: ");
  DatumAusgeben(gd.datum);
  SIO.PutText(", ");
  SIO.PutText(gd.ort);
END GeburtsdatenAusgeben;


PROCEDURE GeschlechtEingeben(): Geschlecht =
(* Eingeben des Geschlechts einer Person *)

VAR ergGeschlecht: Geschlecht := FIRST(Geschlecht);
    c            : CHAR;
    restzeile    : TEXT;
    eingabeOK    : BOOLEAN := FALSE;

BEGIN
  (* Fuehre Abfrage solange durch, bis korrekte Eingabe erfolgt *)
  ZaehleZweig(11);

  REPEAT
    ZaehleZweig(12);
    SIO.PutText("Geschlecht (m, w): ");
    c := SIO.GetChar();
    restzeile := SIO.GetLine(); (* Liest verbleibende Zeichen und RETURN *)
  
    CASE c OF
      'm', 'M' => ZaehleZweig(13);
                  ergGeschlecht := Geschlecht.maennlich;
                  eingabeOK := TRUE;
    | 'w', 'W' => ZaehleZweig(14);
                  ergGeschlecht := Geschlecht.weiblich;
                  eingabeOK := TRUE;
    ELSE
      ZaehleZweig(15);
      SIO.PutLine("Falsche Eingabe!");
    END;
  UNTIL eingabeOK;

  ZaehleZweig(16);
  RETURN ergGeschlecht;
END GeschlechtEingeben;


PROCEDURE GeschlechtAusgeben(g: Geschlecht) =
(* Ausgeben des Geschlechts einer Person *)
BEGIN
  ZaehleZweig(17);

  CASE g OF
    Geschlecht.maennlich => ZaehleZweig(18);
                            SIO.PutText("maennlich");
  | Geschlecht.weiblich  => ZaehleZweig(19);
                            SIO.PutText("weiblich");
  END;
END GeschlechtAusgeben;


PROCEDURE AdresseEingeben(): Adresse =
(* Eingeben einer vollstaendigen Adresse *)
VAR ergAdresse: Adresse;

BEGIN
  ZaehleZweig(20);
  SIO.PutText("Strasse: ");
  ergAdresse.strasse := SIO.GetLine();
  SIO.PutText("Hausnummer: ");
  ergAdresse.nummer := UBEingeben(FIRST(Hausnummer), LAST(Hausnummer));
  SIO.PutText("Zusatz: ");
  ergAdresse.zusatz := SIO.GetLine();
  SIO.PutText("Postleitzahl: ");
  ergAdresse.plz := UBEingeben(FIRST(Postleitzahl), LAST(Postleitzahl));
  SIO.PutText("Ort: ");
  ergAdresse.ort := SIO.GetLine();

  RETURN ergAdresse;
END AdresseEingeben;


PROCEDURE AdresseAusgeben(adr: Adresse) =
(* Ausgeben einer vollstaendigen Adresse *)

BEGIN
  (* Strasse mit Hausnummer und Zusatz in einer Zeile *)
  ZaehleZweig(21);
  SIO.PutText(adr.strasse); SIO.PutText(" "); 
  SIO.PutInt(adr.nummer); SIO.PutText(" ");
  SIO.PutText(adr.zusatz); SIO.Nl();

  (* Postleitzahl mit Ort in neuer Zeile *)
  PostleitzahlAusgeben(adr.plz); SIO.PutText(" ");
  SIO.PutText(adr.ort);
END AdresseAusgeben;


PROCEDURE FamilienstandEingeben(): Familienstand =
(* Eingeben des Familienstands einer Person *)
VAR ergFamilienstand: Familienstand := FIRST(Familienstand);
    c               : CHAR;
    restzeile       : TEXT;
    eingabeOK       : BOOLEAN := FALSE;

BEGIN
  (* Fuehre Abfrage solange durch, bis korrekte Eingabe erfolgt *)
  ZaehleZweig(22);

  REPEAT
    ZaehleZweig(23);
    SIO.PutText("Familienstand (l, v, g, w): ");
    c := SIO.GetChar();
    restzeile := SIO.GetLine(); (* Liest verbleibende Zeichen und RETURN *)
  
    CASE c OF
      'l', 'L' => ZaehleZweig(24);
                  ergFamilienstand := Familienstand.ledig;
                  eingabeOK := TRUE;
    | 'v', 'V' => ZaehleZweig(25);
                  ergFamilienstand := Familienstand.verheiratet;
                  eingabeOK := TRUE;
    | 'g', 'G' => ZaehleZweig(26);
                  ergFamilienstand := Familienstand.geschieden;
                  eingabeOK := TRUE;
    | 'w', 'W' => ZaehleZweig(27);
                  ergFamilienstand := Familienstand.verwitwet;
                  eingabeOK := TRUE;
    ELSE
      ZaehleZweig(28);
      SIO.PutLine("Falsche Eingabe!");
    END;
  UNTIL eingabeOK;

  ZaehleZweig(29);
  RETURN ergFamilienstand;
END FamilienstandEingeben;


PROCEDURE FamilienstandAusgeben(famstand: Familienstand) =
(* Ausgeben des Familienstands einer Person *)

BEGIN
  ZaehleZweig(30);
  CASE famstand OF
    Familienstand.ledig       => ZaehleZweig(31);
                                 SIO.PutText("ledig");
  | Familienstand.verheiratet => ZaehleZweig(32);
                                 SIO.PutText("verheiratet");
  | Familienstand.geschieden  => ZaehleZweig(33);
                                 SIO.PutText("geschieden");
  | Familienstand.verwitwet   => ZaehleZweig(34);
                                 SIO.PutText("verwitwet");
  END;
END FamilienstandAusgeben;


PROCEDURE EinwohnerEingeben(): Einwohner =
(* Komplett Eingeben eines gemeldeten Einwohners *)
VAR ergEinwohner: Einwohner;

BEGIN
  ZaehleZweig(35);
  ergEinwohner.person        := PersonEingeben();
  ergEinwohner.geboren       := GeburtsdatenEingeben();
  ergEinwohner.geschlecht    := GeschlechtEingeben();
  ergEinwohner.wohnung       := AdresseEingeben();
  ergEinwohner.familienstand := FamilienstandEingeben();

  IF ergEinwohner.familienstand = Familienstand.verheiratet THEN
    ZaehleZweig(36);
    SIO.PutLine("Ehepartner: ");
    ergEinwohner.ehepartner := PersonEingeben();
  ELSE
    ZaehleZweig(37);
  END;

  ZaehleZweig(38);
  SIO.PutLine("Anmeldung: ");
  ergEinwohner.anmeldung := DatumEingeben();

  RETURN ergEinwohner;
END EinwohnerEingeben;


PROCEDURE EinwohnerAusgeben(einwohner: Einwohner) =
(* Komplett Ausgeben eines gemeldeten Einwohners *)

BEGIN
  ZaehleZweig(39);
  PersonAusgeben       (einwohner.person); SIO.Nl();
  GeburtsdatenAusgeben (einwohner.geboren); SIO.Nl();
  GeschlechtAusgeben   (einwohner.geschlecht); SIO.Nl();
  SIO.Nl();
  AdresseAusgeben      (einwohner.wohnung); SIO.Nl();
  SIO.Nl();

  (* Familienstand und, falls verheiratet, Ehepartner ausgeben *)
  FamilienstandAusgeben(einwohner.familienstand);
  IF einwohner.familienstand = Familienstand.verheiratet THEN
    ZaehleZweig(40);
    SIO.PutText(" mit ");
    PersonAusgeben(einwohner.ehepartner);
  ELSE
    ZaehleZweig(41);
  END;

  ZaehleZweig(42);
  SIO.Nl();

  (* Datum der Anmeldung ausgeben *)
  SIO.PutText("Gemeldet seit ");
  DatumAusgeben(einwohner.anmeldung); SIO.Nl(); 
END EinwohnerAusgeben;


(* Hilfsprozeduren fuer die Ein- und Ausgabe der Daten *)

PROCEDURE UBEingeben(minimum, maximum: INTEGER): INTEGER =
(* Eingeben einer ganzen Zahl innerhalb des zulaessigen Unterbereichs *)
VAR ergInteger: INTEGER;
    eingabeOK : BOOLEAN := FALSE;

BEGIN
  ZaehleZweig(43);

  (* Fuehre Abfrage solange durch, bis korrekte Eingabe erfolgt *)
  REPEAT
    ZaehleZweig(44);
    ergInteger := SIO.GetInt();
    restzeile  := SIO.GetLine(); (* Liest verbliebene Zeichen und RETURN *)

    IF (ergInteger >= minimum) AND (ergInteger <= maximum) THEN
      (* Wert im zulaessigen Bereich *)
      ZaehleZweig(45);
      eingabeOK := TRUE;
    ELSE
      ZaehleZweig(46);
      SIO.PutText("Wert ungueltig! Bitte nochmal eingeben: ");
    END;
  UNTIL eingabeOK;

  ZaehleZweig(47);
  RETURN ergInteger;
END UBEingeben;


PROCEDURE PostleitzahlAusgeben(plz: [0..99999]) =
(* Hilfsprozedur, um die Postleitzahl immer fuenfstellig auszugeben *)
VAR praefix: TEXT;

BEGIN
  ZaehleZweig(48);

  (* Zu druckendes Praefix bestimmen *)
  CASE plz OF
    0..9         => ZaehleZweig(49);
                    praefix := "0000";
  | 10..99       => ZaehleZweig(50);
                    praefix := "000";
  | 100..999     => ZaehleZweig(51);
                    praefix := "00";
  | 1000..9999   => ZaehleZweig(52);
                    praefix := "0";
  | 10000..99999 => ZaehleZweig(53);
                    praefix := "";
  END;

  (* Erst Praefix, dann eigentliche Zahl ausgeben *)
  ZaehleZweig(54);
  SIO.PutText(praefix);
  SIO.PutInt(plz);
END PostleitzahlAusgeben;


PROCEDURE MelderegisterAusgeben(register: Melderegister; maxIndex: INTEGER) =
(* Gibt das bis zum Eintrag maxIndex gefuellte Melderegister aus *)
BEGIN
  ZaehleZweig(55);

  (* Ausgeben aller Einwohner *)
  FOR i := FIRST(register) TO maxIndex DO
    ZaehleZweig(56);
    SIO.PutInt(i);
    SIO.PutLine(". Einwohner:");
    EinwohnerAusgeben(register[i]);
    SIO.Nl();
  END;

  ZaehleZweig(57);
END MelderegisterAusgeben;


VAR register     : Melderegister;   
    registerIndex: [0..MAXEINWOHNER] := 0;

    (* Notwendig fuer Benutzerinteraktion *)
    auswahl      : CHAR;
    restzeile    : TEXT;             

BEGIN
  (* Benutzerinteraktion mittels REPEAT-Schleife *)
  ZaehleZweig(58);

  REPEAT
    ZaehleZweig(59);
    
    (* Auswahlmenue ausgeben *)
    SIO.Nl();
    SIO.PutLine("***  Einwohnermeldeamt  ***");
    SIO.PutLine("(e) Neuen Einwohner anlegen");
    SIO.PutLine("(a) Alle Einwohner ausgeben");
    SIO.PutLine("(q) Programm beenden");
    SIO.Nl();

    (* Eingabe der ausgewaehlten Operation *)
    SIO.PutText("Auswahl: "); 
    auswahl   := SIO.GetChar();
    restzeile := SIO.GetLine();
    SIO.Nl();

    (* Ausgewaehlte Operation ausfuehren *)
    CASE auswahl OF
       'e', 'E' => ZaehleZweig(60);
                   IF registerIndex < MAXEINWOHNER THEN
                     (* Solange noch Platz im Meldeamt *)
                     ZaehleZweig(61);
                     INC(registerIndex);
                     register[registerIndex] := EinwohnerEingeben();
                   ELSE
                     ZaehleZweig(62);
                     SIO.PutLine("Leider kein Platz mehr frei!");
                   END;
    |  'a', 'A' => ZaehleZweig(63);
                   MelderegisterAusgeben(register, registerIndex);
    |  'q', 'Q' => ZaehleZweig(64);
                   SIO.PutLine("Programmende!");
    ELSE
      ZaehleZweig(65);
      SIO.PutLine("Ungueltige Eingabe!");
    END;
    ZaehleZweig(66);
  UNTIL (auswahl = 'q') OR (auswahl = 'Q');
  ZaehleZweig(67);

  (* Programmende erreicht, schnell noch die Ueberdeckung bestimmen *)
  ZaehlerAuswerten();
END InstMeldeamt.
