MODULE Geldbetrag;

(* Implementierung eines ADTs fuer Geldbetraege.
   Autor           : Antje Nowack
   Umgebung  : PM3, Windows 95
   Erstellt        : 04. 01. 2001  Letzte Aenderung: 09.01. 2001 *)


IMPORT Text;

IMPORT Waehrung; (* Importiere den ADT Waehrung *)

IMPORT SIO;

TYPE REVEAL T = BRANDED REF RECORD
                  wert       : REAL;   
                  waehrung   : Waehrung.T;
                END;

(* Fuer alle folgenden Prozeduren, die auf zwei  Betraegen operieren, ist es
   notwendig, dass die Waehrungen der Betraege gleich sind. *)

PROCEDURE ValidCurrency(waehrung: Waehrung.T) : BOOLEAN =
(* Prueft, ob ein Element vom vom Typ Waehrung.T eine gueltige Waehrung ist. *)
BEGIN
  IF  ((Text.Equal(Waehrung.GibBezeichnung(waehrung), "CHF") AND Waehrung.GibKurs(waehrung) = 1.528781) OR
       (Text.Equal(Waehrung.GibBezeichnung(waehrung), "DM") AND Waehrung.GibKurs(waehrung) = 1.956229) OR
       (Text.Equal(Waehrung.GibBezeichnung(waehrung), "FF") AND Waehrung.GibKurs(waehrung) = 6.560979) OR
       (Text.Equal(Waehrung.GibBezeichnung(waehrung), "EUR")AND Waehrung.GibKurs(waehrung) = 1.00) OR
       (Text.Equal(Waehrung.GibBezeichnung(waehrung), "GBP") AND Waehrung.GibKurs(waehrung) = 0.633604) OR
       (Text.Equal(Waehrung.GibBezeichnung(waehrung), "USD") AND Waehrung.GibKurs(waehrung) = 0.953200))
  THEN RETURN TRUE;
  ELSE RETURN FALSE;
  END;
END ValidCurrency;

PROCEDURE GleicheWaehrung(aBetrag, bBetrag: T) : BOOLEAN =
(* Prueft, ob zwei Betraege die gleiche Waehrung haben. *)
BEGIN 
  IF  Text.Equal(Waehrung.GibBezeichnung(aBetrag^.waehrung), Waehrung.GibBezeichnung(bBetrag^.waehrung))
      AND Waehrung.GibKurs(aBetrag^.waehrung) = Waehrung.GibKurs(bBetrag^.waehrung) 
  THEN RETURN TRUE;
  ELSE RETURN FALSE;
  END;
END GleicheWaehrung;

PROCEDURE Init(VAR betrag: T) =
(* Initialisiert betrag mit NIL. *)
BEGIN
  betrag := NIL;
END Init;

PROCEDURE NoValue(betrag: T) : BOOLEAN =
(* Liefert TRUE gdw. betrag zwar initialisiert ist aber keinen Geldbetrag
   darstellt. *)
BEGIN
  RETURN betrag = NIL;
END NoValue;

PROCEDURE Tausche(betrag: T; waehrung: Waehrung.T): T RAISES{InvalidCurrency, ArgNoValue}=
(* Tauscht den gegebenen Geldbetrag in die angegebene Waehrung und
   gibt den neuen Geldbetrag zurueck. *)

VAR neuerGeldbetrag: T;

BEGIN
  IF NoValue(betrag) THEN RAISE ArgNoValue END;
  IF NOT ValidCurrency(waehrung) THEN RAISE InvalidCurrency END;
  IF NOT ValidCurrency(betrag^.waehrung) THEN RAISE InvalidCurrency END;
  neuerGeldbetrag := NEW(T);
  neuerGeldbetrag^.waehrung := waehrung;
  neuerGeldbetrag^.wert := betrag^.wert * ( Waehrung.GibKurs(waehrung) / Waehrung.GibKurs(betrag^.waehrung) );
  RETURN neuerGeldbetrag;
END Tausche;
  
  
PROCEDURE Erzeuge(skalar: REAL; waehrung: Waehrung.T): T RAISES{InvalidCurrency, NegativeInput}=
(* Erzeugt einen neuen Geldbetrag mit dem gegebenen Wert in der
   angegebenen Waehrung. *)

VAR resGeldbetrag: T;

BEGIN
  IF skalar < 0.0 THEN RAISE NegativeInput END;
  IF NOT ValidCurrency(waehrung) THEN RAISE InvalidCurrency END;
  IF NOT ValidCurrency(waehrung) THEN RAISE InvalidCurrency END;
  resGeldbetrag           := NEW(T);
  resGeldbetrag^.wert     := skalar;
  resGeldbetrag^.waehrung := waehrung;

  RETURN resGeldbetrag;
END Erzeuge;


PROCEDURE Addiere(aBetrag, bBetrag: T): T RAISES{DifferentCurrencies, InvalidCurrency, FirstArgNoValue, SecondArgNoValue}=
(* Addiert den Geldetrag bBetrag zum Betrag aBetrag und gibt das
   Ergebnis in der gemeinsamen Waehrung zurueck. *)

VAR resGeldbetrag: T;

BEGIN
  IF NoValue(aBetrag) THEN RAISE FirstArgNoValue END;
  IF NoValue(bBetrag) THEN RAISE SecondArgNoValue END;
  IF NOT ValidCurrency(aBetrag^.waehrung) THEN RAISE InvalidCurrency END;
  IF NOT ValidCurrency(bBetrag^.waehrung) THEN RAISE InvalidCurrency END;
  IF NOT GleicheWaehrung(aBetrag, bBetrag) THEN RAISE DifferentCurrencies END;
  resGeldbetrag := NEW(T);
  resGeldbetrag^.wert := aBetrag^.wert + bBetrag^.wert;
  resGeldbetrag^.waehrung := aBetrag^.waehrung;

  RETURN resGeldbetrag;
END Addiere;


PROCEDURE Subtrahiere(aBetrag, bBetrag: T): T RAISES{NegativeResult, DifferentCurrencies, InvalidCurrency, FirstArgNoValue, SecondArgNoValue}=
(* Subtrahiert den Geldbetrag bBetrag vom Betrag aBetrag und gibt
   das Ergebnis in der gemeinsamen Waehrung zurueck. Das Ergebnis 
   darf nicht negativ sein. *)

VAR resGeldbetrag: T;

BEGIN
  IF NoValue(aBetrag) THEN RAISE FirstArgNoValue END;
  IF NoValue(bBetrag) THEN RAISE SecondArgNoValue END;
  IF NOT ValidCurrency(aBetrag^.waehrung) THEN RAISE InvalidCurrency END;
  IF NOT ValidCurrency(bBetrag^.waehrung) THEN RAISE InvalidCurrency END;
  IF Kleiner(aBetrag, bBetrag) THEN RAISE NegativeResult END;
  resGeldbetrag := NEW(T);
  resGeldbetrag^.wert := aBetrag^.wert - bBetrag^.wert;
  resGeldbetrag^.waehrung := aBetrag^.waehrung;

  RETURN resGeldbetrag;
END Subtrahiere;


PROCEDURE Multipliziere(betrag: T; skalar: REAL): T RAISES{NegativeResult, InvalidCurrency, ArgNoValue}=
(* Multipliziert den Geldbetrag betrag mit dem gegebenen Skalarwert
   und gibt das Ergebnis in der Waehrung von betrag zurueck. Der Skalarwert
   darf nicht negativ sein. *)

VAR resGeldbetrag: T;

BEGIN
  IF NoValue(betrag) THEN RAISE ArgNoValue END;
  IF NOT ValidCurrency(betrag^.waehrung) THEN RAISE InvalidCurrency END;
  IF skalar < 0.0 THEN RAISE NegativeResult END;
  resGeldbetrag := NEW(T);
  resGeldbetrag^.wert := betrag^.wert * skalar;
  resGeldbetrag^.waehrung := betrag^.waehrung;

  RETURN resGeldbetrag;
END Multipliziere;
  

PROCEDURE Dividiere(betrag: T; skalar: REAL): T RAISES{NegativeResult, InvalidCurrency, DivisionByZero, ArgNoValue}=
(* Dividiert den Geldbetrag betrag durch den gegebenen Skalarwert
   und gibt das Ergebnis in der Waehrung von betrag zurueck. Der Skalarwert
   darf nicht negativ sein und er muss ungleich Null sein. *)

VAR resGeldbetrag: T;

BEGIN
  IF NoValue(betrag) THEN RAISE ArgNoValue END;
  IF NOT ValidCurrency(betrag^.waehrung) THEN RAISE InvalidCurrency END;
  IF skalar = 0.0 THEN RAISE DivisionByZero END;
  IF skalar < 0.0 THEN RAISE NegativeResult END;
  resGeldbetrag := NEW(T);
  resGeldbetrag^.wert := betrag^.wert / skalar;
  resGeldbetrag^.waehrung := betrag^.waehrung;

  RETURN resGeldbetrag;
END Dividiere;

PROCEDURE Print(betrag: T) RAISES{InvalidCurrency, ArgNoValue}=
(* Gibt den Wert und die Waehrungseinheit aus. *)

BEGIN
  IF NoValue(betrag) THEN RAISE ArgNoValue END;
  IF NOT ValidCurrency(betrag^.waehrung) THEN RAISE InvalidCurrency END;
  SIO.PutReal(betrag^.wert);
  SIO.PutLine(Waehrung.GibBezeichnung(betrag^.waehrung));
END Print;


PROCEDURE Gleich(aBetrag, bBetrag: T): BOOLEAN RAISES{DifferentCurrencies, InvalidCurrency, FirstArgNoValue, SecondArgNoValue}=
(* Prueft, ob die beiden Geldbetraege gleich gross sind. *)
BEGIN
  IF NoValue(aBetrag) THEN RAISE FirstArgNoValue END;
  IF NoValue(bBetrag) THEN RAISE SecondArgNoValue END;
  IF NOT ValidCurrency(aBetrag^.waehrung) THEN RAISE InvalidCurrency END;
  IF NOT ValidCurrency(bBetrag^.waehrung) THEN RAISE InvalidCurrency END;
  IF NOT GleicheWaehrung(aBetrag, bBetrag) THEN RAISE DifferentCurrencies END;
  RETURN aBetrag^.wert = bBetrag^.wert;
END Gleich;


PROCEDURE Kleiner(aBetrag, bBetrag: T): BOOLEAN RAISES{DifferentCurrencies, InvalidCurrency, FirstArgNoValue, SecondArgNoValue}=
(* Prueft, ob der Geldbetrag aBetrag kleiner als bBetrag ist. *)
BEGIN
  IF NoValue(aBetrag) THEN RAISE FirstArgNoValue END;
  IF NoValue(bBetrag) THEN RAISE SecondArgNoValue END;
  IF NOT ValidCurrency(aBetrag^.waehrung) THEN RAISE InvalidCurrency END;
  IF NOT ValidCurrency(bBetrag^.waehrung) THEN RAISE InvalidCurrency END;
  IF NOT GleicheWaehrung(aBetrag, bBetrag) THEN RAISE DifferentCurrencies END;
  RETURN aBetrag^.wert < bBetrag^.wert;
END Kleiner;


PROCEDURE Groesser(aBetrag, bBetrag: T): BOOLEAN RAISES{DifferentCurrencies, InvalidCurrency, FirstArgNoValue, SecondArgNoValue}=
(* Prueft, ob der Geldbetrag aBetrag groesser als bBetrag ist. *)
BEGIN
  IF NoValue(aBetrag) THEN RAISE FirstArgNoValue END;
  IF NoValue(bBetrag) THEN RAISE SecondArgNoValue END;
  IF NOT ValidCurrency(aBetrag^.waehrung) THEN RAISE InvalidCurrency END;
  IF NOT ValidCurrency(bBetrag^.waehrung) THEN RAISE InvalidCurrency END;
  IF NOT GleicheWaehrung(aBetrag, bBetrag) THEN RAISE DifferentCurrencies END;
  RETURN aBetrag^.wert > bBetrag^.wert;
END Groesser;


BEGIN
END Geldbetrag.
